unit UCalc;

interface

uses
  Classes, SysUtils, Dialogs, ComCtrls;

type
  TArray = array[1..6] of Integer;

  TOp = record
    A, B: Integer;
    Op: Char;
    Res: Integer;
  end;

  TAOp = array[1..5] of TOp;

  TPOps = ^TOps;
  TOps = object
    AOp: TAOp;
    NbOp: Byte;
    constructor Create;
    procedure AddOp(Calc: TOps; _A, _B: Integer; _Op: Char);
  end;

  TSeq = object
    Nbs: array[1..6] of Integer;
    Count: Integer;
    Op: Char;
    Res: Integer;
    function Equals(s2: TSeq): Boolean;
    function OpCompatibles(Op2: Char): Boolean;
    procedure Init;
  end;

  TSeqs = object
    ASeq: array[1..5] of TSeq;
    NbSeq: Integer;
    procedure Init;
    procedure AddOp(Op: TOp);
    function Equals(S2: TSeqs): Boolean;
  end;

  TCalcThread = class(TThread)
  Private
    FT: TArray;
    FRes: Integer;
    FListe: TStringList;
    FPb: TProgressBar;
    procedure ChercheSolutions;
  Protected
    procedure UpdateProgressBar;
    procedure FullProgressBar;
    procedure Execute; Override;
  Public
    constructor Create(T: TArray; Res: Integer; List: TStringList; Pb: TProgressBar);
  end;

implementation

//--------------------------------------------------

constructor TOps.Create;
var
  I: Integer;
begin
  NbOp := 0;
  for I := 1 to 5 do
  begin
    AOp[I].A := -1;
    AOp[I].B := -1;
    AOp[I].Op := '_';
    AOp[I].Res := -1;
  end;
end;

//--------------------------------------------------

procedure TOps.AddOp(Calc: TOps; _A, _B: Integer; _Op: Char);
begin
  NbOp := Calc.NbOp + 1;
  AOp := Calc.AOp;
  if not (NbOp in [1..5]) then Exit;
  AOp[NbOp].A := _A;
  AOp[NbOp].B := _B;
  AOp[NbOp].Op := _Op;
end;

//--------------------------------------------------

function TSeq.Equals(s2: TSeq): Boolean;
var
  I, J: Integer;
  NbMemeNb: Integer;
begin
  NbMemeNb := 0;
  if (not (OpCompatibles(S2.Op)) or (Res <> s2.Res) or (Count <> s2.Count) or (Res = -1)) then
    Result := False
  else
  begin
    for I := 1 to Count do
      for j := 1 to Count do
        if (Nbs[I] = s2.Nbs[J]) then
        begin
          s2.Nbs[J] := -1;
          Inc(NbMemeNb);
          Break;
        end;
    Result := (NbMemeNb = Count);
  end;
end;

//--------------------------------------------------

function TSeq.OpCompatibles(Op2: Char): Boolean;
begin
  Result := ((Op in ['+', '-']) and (Op2 in ['+', '-']))
         or ((Op in ['*', '/']) and (Op2 in ['*', '/']));
end;

//--------------------------------------------------

procedure TSeq.Init;
var
  I: Integer;
begin
  for I := 1 to 6 do
    Nbs[I] := 0;
  Res := -1;
  Op := '_';
  Count := 0;
end;

//--------------------------------------------------

procedure TSeqs.Init;
var
  I: Integer;
begin
  for I := 1 to 5 do
    Self.ASeq[I].Init;
  Self.NbSeq := 0;
end;

//--------------------------------------------------

procedure TSeqs.AddOp(Op: TOp);
var
  I, J: Integer;
begin
  for I := 1 to NbSeq do
    if (ASeq[I].OpCompatibles(Op.Op)) then
      for J := 1 to ASeq[I].Count do
        if (Op.Res = ASeq[I].Nbs[J]) then
        begin
          ASeq[I].Nbs[J] := Op.A;
          Inc(ASeq[I].Count);
          ASeq[I].Nbs[ASeq[I].Count] := Op.B;
          Exit;
        end;
  Inc(NbSeq);
  ASeq[NbSeq].Nbs[1] := Op.A;
  ASeq[NbSeq].Nbs[2] := Op.B;
  ASeq[NbSeq].Count := 2;
  ASeq[NbSeq].Op := Op.Op;
  ASeq[NbSeq].Res := Op.Res;
end;

//--------------------------------------------------

function TSeqs.Equals(S2: TSeqs): Boolean;
var
  I, J: Integer;
  NbMemeSeq: Integer;
  Marques: array[1..5] of Boolean;
begin
  NbMemeSeq := 0;
  for I := 1 to 5 do
    Marques[I] := False;
  for I := 1 to NbSeq do
    for J := 1 to S2.NbSeq do
      if (not(Marques[J]) and ASeq[I].Equals(S2.ASeq[J])) then
      begin
        Inc(NbMemeSeq);
        Marques[J] := True;
      end;
  Result := (NbMemeSeq = NbSeq);
end;

//--------------------------------------------------

constructor TCalcThread.Create(T: TArray; Res: Integer; List: TStringList; Pb: TProgressBar);
begin
  inherited Create(False);
  FT := T;
  FRes := Res;
  FListe := List;
  FListe.Clear;
  Priority := tpLower;
  FPb := Pb;
  FPb.Position := 0;
end;

//--------------------------------------------------

procedure TCalcThread.ChercheSolutions;
var
  OpList: TList;
  UnCalc: TOps;
  NearestRes1, NearestRes2: Integer;
  InfoStr: string;

  //------------------------------------------------
  procedure AddCompte(Calc: TOps);
  var
    I, J: Integer;
    Calc1, Calc2: TOps;
    ptInt: TPOps;
    Trouve: Boolean;
    Seqs1, seqs2: TSeqs;

  begin //AddCompte
    //on verifie que chaque resultat intermediaire est utilise plus loin
    for I := 1 to Calc.NbOp do
    begin
      case Calc.AOp[I].Op of
        '+': Calc.AOp[I].Res := Calc.AOp[I].A + Calc.AOp[I].B;
        '-': Calc.AOp[I].Res := Calc.AOp[I].A - Calc.AOp[I].B;
        '*': Calc.AOp[I].Res := Calc.AOp[I].A * Calc.AOp[I].B;
        '/': Calc.AOp[I].Res := Calc.AOp[I].A div Calc.AOp[I].B;
      else Calc.AOp[I].Res := -1;
      end;
      Trouve := (I = Calc.NbOp);
      for J := I + 1 to Calc.NbOp do
        Trouve := Trouve or (Calc.AOp[J].A = Calc.AOp[I].Res) or (Calc.AOp[J].B = Calc.AOp[I].Res);
      if not (Trouve) then Exit;
    end;
    //on verifie que la solution n'est pas deja repertoriee
    Calc1 := Calc;
    Seqs1.Init;
    for I := Calc1.NbOp downto 1 do
      Seqs1.AddOp(Calc1.AOp[I]);
    for I := 0 to OpList.Count - 1 do
    begin
      Calc1 := Calc;
      Calc2 := TPOps(OpList[I])^;
      Seqs2.Init;
      if (Calc2.NbOp = Calc1.NbOp) then
      begin
        for J := Calc2.NbOp downto 1 do
          Seqs2.AddOp(Calc2.AOp[J]);
        if (Seqs2.Equals(Seqs1)) then Exit;
      end;
    end;
    new(ptInt);
    ptInt^ := Calc;
    OpList.Add(ptInt);
  end; //AddCompte

  //------------------------------------------------
  procedure Calcul(T: TArray; Calc: TOps; profond: Integer); //procedure recursive
  var
    Calcs: TOps;
    T2: TArray;
    I, J, Res: Integer;
  begin //Calcul
    Calcs := Calc;
    if not (Terminated) then
      for I := 1 to 6 do
        if (T[I] <> 0) then
        begin

          //T[I] + T[J]
          for J := I + 1 to 6 do
          begin
            if (profond = 0) then Synchronize(UpdateProgressBar);
            if (T[J] <> 0) then
            begin
              Res := T[I] + T[J];
              Calcs.AddOp(Calc, T[I], T[J], '+');
              if (Res = FRes) then
              begin
                AddCompte(Calcs);
                Exit;
              end;
              if (Abs(FRes - Res) < Abs(FRes - NearestRes1)) then
              begin
                NearestRes1 := Res;
                Nearestres2 := 0;
              end
              else if ((Res <> NearestRes1) and (Abs(FRes - Res) = Abs(FRes - NearestRes1))) then
                NearestRes2 := Res;
              T2 := T;
              T2[J] := 0;
              T2[I] := Res;
              Calcul(T2, Calcs, 1);
            end;
          end;

          //T[I] * T[J]
          for J := I + 1 to 6 do
          begin
            if (profond = 0) then Synchronize(UpdateProgressBar);
            if (T[J] <> 0) then
            begin
              if ((T[I] <> 1) and (T[J] <> 1)) then //on evite le 1*_=_
              begin
                Res := T[I] * T[J];
                Calcs.AddOp(Calc, T[I], T[J], '*');
                if (Res = FRes) then
                begin
                  AddCompte(Calcs);
                  Exit;
                end;
                if (Abs(FRes - Res) < Abs(FRes - NearestRes1)) then
                begin
                  NearestRes1 := Res;
                  Nearestres2 := 0;
                end
                else if ((Res <> NearestRes1) and (Abs(FRes - Res) = Abs(FRes - NearestRes1))) then
                  NearestRes2 := Res;
                T2 := T;
                T2[J] := 0;
                T2[I] := Res;
                Calcul(T2, Calcs, 1);
              end;
            end;
          end;

          //T[I] - T[J]
          for J := I + 1 to 6 do
          begin
            if (profond = 0) then Synchronize(UpdateProgressBar);
            if ((T[J] <> 0) and (T[J] <> T[I])) then
            begin
              if (T[I] > T[J]) then
              begin
                Res := T[I] - T[J]; //T[I] - T[J] > 0
                if (Res <> T[J]) then //on evite le 4-2=2
                begin
                  Calcs.AddOp(Calc, T[I], T[J], '-');
                  if (Res = FRes) then
                  begin
                    AddCompte(Calcs);
                    Exit;
                  end;
                  if (Abs(FRes - Res) < Abs(FRes - NearestRes1)) then
                  begin
                    NearestRes1 := Res;
                    Nearestres2 := 0;
                  end
                  else if ((Res <> NearestRes1) and (Abs(FRes - Res) = Abs(FRes - NearestRes1))) then
                    NearestRes2 := Res;
                  T2 := T;
                  T2[J] := 0;
                  T2[I] := Res;
                  Calcul(T2, Calcs, 1);
                end;
              end
              else begin
                Res := T[J] - T[I]; //T[J] - T[I] > 0
                if (Res <> T[J]) then //on evite le 4-2=2
                begin
                  Calcs.AddOp(Calc, T[J], T[I], '-');
                  if (Res = FRes) then
                  begin
                    AddCompte(Calcs);
                    Exit;
                  end;
                  if (Abs(FRes - Res) < Abs(FRes - NearestRes1)) then
                  begin
                    NearestRes1 := Res;
                    Nearestres2 := 0;
                  end
                  else if ((Res <> NearestRes1) and (Abs(FRes - Res) = Abs(FRes - NearestRes1))) then
                    NearestRes2 := Res;
                  T2 := T;
                  T2[J] := 0;
                  T2[I] := Res;
                  Calcul(T2, Calcs, 1);
                end;
              end;
            end;
          end;

          //T[I] / T[J]
          for J := I + 1 to 6 do
          begin
            if (profond = 0) then Synchronize(UpdateProgressBar);
            if (T[J] <> 0) then
            begin
              if (T[I] > T[J]) then
              begin
                Res := T[I] div T[J]; //T[I] div T[J] >= 1
                if ((T[I] mod T[J] = 0) and (T[J] <> 1) and (T[J] <> Res)) then //on evite le _/1=_ et le 4/2=2
                begin
                  Calcs.AddOp(Calc, T[I], T[J], '/');
                  if (Res = FRes) then
                  begin
                    AddCompte(Calcs);
                    Exit;
                  end;
                  if (Abs(FRes - Res) < Abs(FRes - NearestRes1)) then
                  begin
                    NearestRes1 := Res;
                    Nearestres2 := 0;
                  end
                  else if ((Res <> NearestRes1) and (Abs(FRes - Res) = Abs(FRes - NearestRes1))) then
                    NearestRes2 := Res;
                  T2 := T;
                  T2[J] := 0;
                  T2[I] := Res;
                  Calcul(T2, Calcs, 1);
                end;
              end
              else begin
                Res := T[J] div T[I]; //T[I] div T[J] >= 1
                if ((T[J] mod T[I] = 0) and (T[I] <> 1) and (T[I] <> Res)) then //on evite le _/1=_ et le 4/2=2
                begin
                  Calcs.AddOp(Calc, T[J], T[I], '/');
                  if (Res = FRes) then
                  begin
                    AddCompte(Calcs);
                    Exit;
                  end;
                  if (Abs(FRes - Res) < Abs(FRes - NearestRes1)) then
                  begin
                    NearestRes1 := Res;
                    Nearestres2 := 0;
                  end
                  else if ((Res <> NearestRes1) and (Abs(FRes - Res) = Abs(FRes - NearestRes1))) then
                    NearestRes2 := Res;
                  T2 := T;
                  T2[J] := 0;
                  T2[I] := Res;
                  Calcul(T2, Calcs, 1);
                end;
              end;
            end;
          end;
        end;
  end; //Calcul

  //------------------------------------------------
  procedure Trier;
  var
    I: Integer;
    RgMinCalc: Integer;

    //----------------------------------------------
    function ResMax(c: TOps): Integer;
    var
      K: Integer;
      Res: Integer;
    begin //ResMax
      Res := c.AOp[1].Res;
      for K := 2 to c.NbOp do
        if (c.AOp[K].Res > Res) then Res := c.AOp[K].Res;
      Result := Res;
    end; //ResMax

    //----------------------------------------------
    function OpsToStr(c: TOps): string;
    var
      I: Integer;
      Res: string;
    begin //OpsToStr
      Res := InfoStr;
      for I := 1 to c.NbOp do
        Res := Res + IntToStr(c.AOp[I].A) + ' ' + c.AOp[I].Op + ' ' + IntToStr(c.AOp[I].B) + ' = ' + IntToStr(c.AOp[I].Res) + Chr(13);
      Result := Res;
    end; //OpsToStr

  begin //Trier
    while (OpList.Count > 0) do
    begin
      RgMinCalc := 0;
      for I := 1 to OpList.Count - 1 do
        if ((TPOps(OpList[I])^.NbOp < TPOps(OpList[RgMincalc])^.NbOp)
          or (ResMax(TPOps(OpList[I])^) < ResMax(TPOps(OpList[RgMinCalc])^))) then
          RgMinCalc := I;
      FListe.Add(OpsToStr(TPOps(OpList[RgMincalc])^));
      Dispose(OpList[RgMinCalc]);
      OpList.Delete(RgMinCalc);
    end;
  end; //Trier

begin //ChercheSolutions
  if (FListe = nil) then FListe := TStringList.Create;
  OpList := TList.Create;
  UnCalc.Create;
  NearestRes1 := 0;
  NearestRes2 := 0;
  InfoStr := '';
  Calcul(FT, UnCalc, 0);
  if (Oplist.Count = 0) then
  begin
    if ((NearestRes2 > 0) and (NearestRes2 < NearestRes1)) then
    begin
      FRes := NearestRes2;
      NearestRes2 := NearestRes1;
      NearestRes1 := FRes;
    end
    else FRes := NearestRes1;
    UnCalc.Create;
    Calcul(FT, UnCalc, 0);
    if (NearestRes2 <= 0) then
      InfoStr := 'No solution. Nearest result is ' + IntToStr(NearestRes1) + Chr(13)
    else begin
      InfoStr := 'No solution. Nearest results are ' + IntToStr(NearestRes1) + ' et ' + IntToStr(NearestRes2) + Chr(13);
      FRes := NearestRes2;
      UnCalc.Create;
      Calcul(FT, UnCalc, 0);
    end;
  end;
  Trier;
  OpList.Free;
  Synchronize(FullProgressBar);
end; //ChercheSolutions

//--------------------------------------------------

procedure TCalcThread.UpdateProgressBar;
begin
  FPb.StepIt;
end;

//--------------------------------------------------

procedure TCalcThread.FullProgressBar;
begin
  FPb.Position := FPb.Max;
end;

//--------------------------------------------------

procedure TCalcThread.Execute;
begin
  ChercheSolutions;
end;

//--------------------------------------------------

end.

